/*
 * Multichat Project
 * Base Protocol Manager
 * baseconnection.java -- handles the Base protocol on a low level
 * Copyright (C) 1999 z2 Communications, Inc.
 * Author: Filip Pizlo
 *
 * Heavily hacked by now.  Base will:
 * -> no longer get confused when key max is exceeded
 * -> no longer attempt to send or recieve packets of obscene size
 * -> run quite a bit faster
 * -> more efficiently handle those large packets
 *
 * Definition of terms:
 * packet            - any single entity transmitted one way or the other
 * clear packet      - packet which indicates that the incoming id database
 *                     should be discarded
 * define packet     - defines an id/key pair for the incoming id database
 * data packet       - packet containing id/value pairs, where ids are
 *                     translated into their respective keys using the incoming
 *                     id database
 * inline key packet - packet containing key/value pairs; this is slower, but
 *                     allows for much larger packets
 */

import java.io.*;
import java.util.*;
import java.net.*;

public class baseconnection {
 private DataInputStream fin;
 private DataOutputStream fout;
 private Hashtable 
  n2i,	// this handles outgoing ids
  i2n;	// this handles incoming ids
 private short cnt=0;
 private baseconnection() {
  n2i=new Hashtable();
  i2n=new Hashtable();
 }
 public baseconnection(DataInputStream _fin,DataOutputStream _fout) {
  this();
  fin=_fin;
  fout=_fout;
 }
 public baseconnection(InputStream in,OutputStream out) {
  this(new DataInputStream(in),new DataOutputStream(out));
 }
 public baseconnection(Socket sock) throws IOException {
  this(sock.getInputStream(),sock.getOutputStream());
 }
 // damned Java can't have unsigned variables....
 private int _32to8(int x) throws IOException {
  if (x>=0&&x<=127) return x;
  if (x>=128&&x<=255) return -(x-127);
  throw new IOException("Bad number");
 }
 private int _8to32(int x) {
  if (x<0) return (-x)+127;
  return x;
 }
 // our own String writer/reader; this assumes that the stream already has
 // the data available.  anotherwords, the supplied streams must be from/to
 // ByteStreams.
 private void writeString(DataOutputStream ofout,String str)
  throws IOException {
  ofout.write(_32to8(str.length()));
  ofout.write(str.getBytes());
 }
 private String readString(DataInputStream ofin) throws IOException {
  int l=ofin.read();
  l=_8to32(l);
  byte[] data=new byte[l];
  ofin.read(data);
  return new String(data);
 }
 // boring settings
 private int packmax=262144,maxtokens=8192;
 public void setpackmax(int _packmax) { packmax=_packmax; }
 public void setmaxtokens(int _maxtokens) { maxtokens=_maxtokens; }
 // NOW, lets move on the FUN STUFF!
 private final synchronized void sendpacket(byte[] data) throws IOException {
  if (data.length>packmax) return;
  fout.writeInt(data.length);
  fout.write(data);
  sent++;
  int curpack=data.length+4;
  bsent+=curpack;
  if (curpack>maxpacksent) maxpacksent=curpack;
 }
 private final byte[] recievepacket() throws IOException {
  int torec=fin.readInt();
  brecieved+=4;
  int curpack=4;
  if (torec>packmax) {
   fin.skip(torec);
   return null;
  }
  ByteArrayOutputStream bout=new ByteArrayOutputStream();
  byte[] relay=new byte[65536];
  int recnow=torec>relay.length?relay.length:torec;
  while (torec>0) {
   int res=fin.read(relay,0,recnow);
   if (res<=0) throw new IOException("Disconnected");
   torec-=res;
   recnow=torec>relay.length?relay.length:torec;
   bout.write(relay,0,res);
  }
  byte[] ret=bout.toByteArray();
  recieved++;
  brecieved+=ret.length;
  curpack+=ret.length;
  if (curpack>maxpackrecieved) maxpackrecieved=curpack;
  return ret;
 }
 private final short getnewid() {
  return cnt++;
 }
 private final void assignid(short id,String name) throws IOException {
  n2i.put(name,new Short(id));
  ByteArrayOutputStream bout=new ByteArrayOutputStream();
  DataOutputStream fbout=new DataOutputStream(bout);
  fbout.writeShort(0);
  fbout.writeShort(id);
  writeString(fbout,name);
  sendpacket(bout.toByteArray());
 }
 int resets=0;
 private final void clear() throws IOException {
  resets++;
  n2i.clear();
  cnt=0;
  ByteArrayOutputStream bout=new ByteArrayOutputStream();
  DataOutputStream fbout=new DataOutputStream(bout);
  fbout.writeShort(-1);
  sendpacket(bout.toByteArray());
 }
 public final boolean checksize(Hashtable huh) throws IOException {
  int size=huh.size();
  if (size>maxtokens) return false;
  if (size>65536-cnt) clear();
  return true;
 }
 long
  sent=0,recieved=0,bsent=0,brecieved=0,maxpacksent=0,maxpackrecieved=0,
  sentinline=0;
 public final Properties getstats() {
  Properties ret=new Properties();
  ret.put("outgoing.tokens",""+n2i.size());
  ret.put("incoming.tokens",""+i2n.size());
  ret.put("bytes.sent",""+bsent);
  ret.put("bytes.recieved",""+brecieved);
  ret.put("packets.sent",""+sent);
  ret.put("packets.recieved",""+recieved);
  ret.put("largest.packet.sent",""+maxpacksent);
  ret.put("largest.packet.recieved",""+maxpackrecieved);
  ret.put("reset.n",""+resets);
  ret.put("inline.key.packets.sent",""+sentinline);
  return ret;
 }
 public final synchronized void inlinesend(Hashtable table)
  throws IOException {
  // this sends an inline packet
  int sumofdata=0;
  byte[] data;
  ByteArrayOutputStream bout=new ByteArrayOutputStream();
  DataOutputStream fbout=new DataOutputStream(bout);
  Enumeration enum=table.keys();
  int numtokens=0;
  while (enum.hasMoreElements()) {
   String tmp=(String)enum.nextElement();
   try {
    data=(byte[])table.get(tmp);
   } catch (Throwable e) {
    // convert the data first to a String and then to a byte[]
    String tmp2=table.get(tmp).toString();
    data=tmp2.getBytes();
   }
   // sumofdata is only a ROUGH ESTIMATE of the size, so we can catch those
   // obscenely large packets before we spend too much RAM and processing time
   // on them.  the final verdict will be called in the sendpacket method.
   sumofdata+=data.length+tmp.length();
   if (sumofdata>packmax) return;
   writeString(fbout,tmp);
   fbout.writeInt(data.length);
   fbout.write(data);
   numtokens++;
  }
  data=bout.toByteArray();
  bout=new ByteArrayOutputStream(2+data.length);
  fbout=new DataOutputStream(bout);
  fbout.writeShort(-2);	// indicate inline key packet
  fbout.writeInt(numtokens);
//  System.out.println("inlinesend: numtokens="+numtokens);
  fbout.write(data);
  sendpacket(bout.toByteArray());
  sentinline++;
 }
 public final synchronized void send(Hashtable table) throws IOException {
  // if there is more keys in this table than we can handle, start over or
  // use an inline packet
  boolean traditional=checksize(table);
  if (!traditional) {
   inlinesend(table);
   return;
  }
  // send the data packet, possibly sending define packets first
  int sumofdata=0;
  byte[] data;
  ByteArrayOutputStream bout=new ByteArrayOutputStream();
  DataOutputStream fbout=new DataOutputStream(bout);
  Enumeration enum=table.keys();
  int numtokens=0;
  while (enum.hasMoreElements()) {
   String tmp=(String)enum.nextElement();
   short id=-1;
   Object o=n2i.get(tmp);
   if (o==null) {
    // send a define packet
    id=getnewid();
    assignid(id,tmp);
   } else id=((Short)o).shortValue();
   try {
    data=(byte[])table.get(tmp);
   } catch (Throwable e) {
    // convert the data first to a String and then to a byte[]
    String tmp2=table.get(tmp).toString();
    data=tmp2.getBytes();
   }
   // sumofdata is only a ROUGH ESTIMATE of the size, so we can catch those
   // obscenely large packets before we spend too much RAM and processing time
   // on them.  the final verdict will be called in the sendpacket method.
   sumofdata+=data.length;
   if (sumofdata>packmax) return;
   fbout.writeShort(id);
   fbout.writeInt(data.length);
   fbout.write(data);
   numtokens++;
  }
  data=bout.toByteArray();
  bout=new ByteArrayOutputStream(2+data.length);
  fbout=new DataOutputStream(bout);
//  System.out.println("send: numtokens="+numtokens);
  fbout.writeShort(numtokens);
  fbout.write(data);
  sendpacket(bout.toByteArray());
 }
 public final Hashtable recieve() throws IOException {
  Hashtable ret=new Hashtable();
  while (true) {
   ByteArrayInputStream bin=new ByteArrayInputStream(recievepacket());
   DataInputStream fbin=new DataInputStream(bin);
   // lng is the length in tokens (if positive) or packet type (if negative
   // or zero)
   int lng=fbin.readShort();
   if (lng==-1) {	// lng=-1 indicates clear packet
    i2n.clear();
   } else if (lng==-2) {	// lng=-2 indicates inline key packet
    int lng2=fbin.readInt();
    while (lng2>0) {
     String key=readString(fbin);
     int nl=fbin.readInt();
     byte[] data=new byte[nl];
     fbin.read(data);
     lng--;
     ret.put(key,data);
    }
    break;
   } else if (lng==0) {	// lng=0 indicates define packet
    short id=fbin.readShort();
    String tmp=readString(fbin);
    i2n.put(new Short(id),tmp);
   } else if (lng>0) {	// positive length indicates data packet
    while (lng>0) {
     short id=fbin.readShort();
     int nl=fbin.readInt();
     byte[] data=new byte[nl];
     fbin.read(data);
     lng--;
     Object o=i2n.get(new Short(id));
     ret.put(o.toString(),data);
    }
    break;
   }
  }
  return ret;
 }
}




